// SPDX-License-Identifier: GPL-2.0-or-later
/*
 *
 * Copyright (C) 2021 CERN (www.cern.ch)
 *
 * Released according to the GNU GPL, version 2 or any later version
 *
 * Driver for SVEC (Simple VME FMC carrier) board.
 */

#include <linux/bitmap.h>
#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/firmware.h>
#include <linux/fmc.h>
#include <linux/fs.h>
#include <linux/io.h>
#include <linux/jhash.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/dma-mapping.h>
#include <vmebus.h>

#include "svec.h"

#define SVEC_DMA_DDR_ALIGN 4

/**
 * dma_cookie_complete - complete a descriptor
 * @tx: descriptor to complete
 *
 * Mark this descriptor complete by updating the channels completed
 * cookie marker.  Zero the descriptors cookie to prevent accidental
 * repeated completions.
 *
 * Note: caller is expected to hold a lock to prevent concurrency.
 */
static inline void dma_cookie_complete(struct dma_async_tx_descriptor *tx)
{
	BUG_ON(tx->cookie < DMA_MIN_COOKIE);
	tx->chan->completed_cookie = tx->cookie;
	tx->cookie = 0;
}

/**
 * dma_cookie_init - initialize the cookies for a DMA channel
 * @chan: dma channel to initialize
 */
static inline void dma_cookie_init(struct dma_chan *chan)
{
	chan->cookie = DMA_MIN_COOKIE;
	chan->completed_cookie = DMA_MIN_COOKIE;
}

/**
 * dma_cookie_assign - assign a DMA engine cookie to the descriptor
 * @tx: descriptor needing cookie
 *
 * Assign a unique non-zero per-channel cookie to the descriptor.
 * Note: caller is expected to hold a lock to prevent concurrency.
 */
static inline dma_cookie_t dma_cookie_assign(struct dma_async_tx_descriptor *tx)
{
	struct dma_chan *chan = tx->chan;
	dma_cookie_t cookie;

	cookie = chan->cookie + 1;
	if (cookie < DMA_MIN_COOKIE)
		cookie = DMA_MIN_COOKIE;
	tx->cookie = chan->cookie = cookie;

	return cookie;
}

static inline struct svec_dma_chan *to_svec_dma_chan(struct dma_chan *_ptr)
{
	return container_of(_ptr, struct svec_dma_chan, chan);
}

static int svec_dma_alloc_chan_resources(struct dma_chan *dchan)
{
	return 0;
}

static void svec_dma_free_chan_resources(struct dma_chan *dchan)
{
}

static int svec_dma_slave_config(struct dma_chan *dchan,
                                 struct dma_slave_config *sconfig)
{
	struct svec_dma_chan *chan = to_svec_dma_chan(dchan);

        /* Source address must be word aligned */
	if (sconfig->src_addr & 3)
		return -EINVAL;

	memcpy(&chan->sconfig, sconfig, sizeof(struct dma_slave_config));

	return 0;
}

/**
 * DMA transfer descriptor
 * @tx: dmaengine descriptor
 * @sgl_hw: scattelist HW descriptors
 * @sg_len: number of entries in the scatterlist
 * @direction: transfer direction
 * @vaddr: VME source/destination address
 * @list: the entry is added to the channel list
 */
struct svec_dma_tx {
	struct dma_async_tx_descriptor tx;
        struct scatterlist *sgl;
	unsigned int sg_len;
        enum dma_transfer_direction direction;
        dma_addr_t vaddr;
	struct list_head list;

	struct dma_async_tx_descriptor *vtx;
	dma_cookie_t cookie;
};

static inline struct svec_dma_tx *to_svec_dma_tx(struct dma_async_tx_descriptor *_ptr)
{
	return container_of(_ptr, struct svec_dma_tx, tx);
}

static int svec_dma_terminate_all(struct dma_chan *dchan)
{
	struct svec_dma_chan *chan = to_svec_dma_chan(dchan);
	struct svec_fpga *svec_fpga = container_of(dchan->device, struct svec_fpga, dma);
        unsigned long flags;
        unsigned long flags2;
        struct svec_dma_tx *tx, *tx_tmp;

        const struct dmaengine_result result = {
                .result = DMA_TRANS_ABORTED,
                .residue = 0,
        };

        spin_lock_irqsave(&svec_fpga->dma_lock, flags);

        /* Abort pending queued requests */
        list_for_each_entry_safe(tx, tx_tmp,
				 &svec_fpga->dma_pending_list, list) {
                if (tx->tx.chan == dchan) {
                        list_del(&tx->list);
                        tx->tx.callback_result(tx->tx.callback_param, &result);
                        kfree(tx);
                }
        }

        /* Abort submitted requests */
        spin_lock_irqsave(&chan->lock, flags2);
        list_for_each_entry_safe(tx, tx_tmp, &chan->pending_list, list) {
                list_del(&tx->list);
                kfree(tx);
        }
	spin_unlock_irqrestore(&chan->lock, flags2);

        spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);

        /* Leave the current tx as is. It may not be part of the channel,
           and it is not easy to deal with it */
	return 0;
}

/**
 * Add a descriptor to the pending DMA transfer queue.
 * This will not trigger any DMA transfer: here we just collect DMA
 * transfer descriptions.
 */
static dma_cookie_t svec_dma_tx_submit(struct dma_async_tx_descriptor *tx)
{
	struct svec_dma_tx *svec_dma_tx = to_svec_dma_tx(tx);
	struct svec_dma_chan *chan = to_svec_dma_chan(tx->chan);
	dma_cookie_t cookie;
	unsigned long flags;

	dev_dbg(&tx->chan->dev->device, "%s submit %p\n", __func__, tx);

        /* Allocate cookie and append the tx descriptor */
	spin_lock_irqsave(&chan->lock, flags);
	cookie = dma_cookie_assign(tx);
	list_add_tail(&svec_dma_tx->list, &chan->pending_list);
	spin_unlock_irqrestore(&chan->lock, flags);

	return cookie;
}

static struct dma_async_tx_descriptor *svec_dma_prep_slave_sg(
	struct dma_chan *chan, struct scatterlist *sgl, unsigned int sg_len,
	enum dma_transfer_direction direction, unsigned long flags,
	void *context)
{
	/* struct dma_device *dma = chan->device; */
	struct dma_slave_config *sconfig = &to_svec_dma_chan(chan)->sconfig;
	struct svec_dma_tx *svec_dma_tx;
	struct scatterlist *sg;
	int i;

	if (unlikely(sconfig->direction != direction)) {
		dev_err(&chan->dev->device,
			"Transfer and slave configuration disagree on DMA direction\n");
                return NULL;
	}

	if (unlikely(!sgl || !sg_len)) {
		dev_err(&chan->dev->device,
			"You must provide a DMA scatterlist\n");
                return NULL;
	}

        for_each_sg(sgl, sg, sg_len, i) {
                if (sg_dma_len(sg) > dma_get_max_seg_size(chan->device->dev)) {
			dev_err(&chan->dev->device,
				"Maximum transfer size %d, got %d on transfer %d\n",
				dma_get_max_seg_size(chan->device->dev),
				sg_dma_len(sg), i);
                        return NULL;
		}
		if (sg_dma_len(sg) & (SVEC_DMA_DDR_ALIGN - 1)) {
			dev_err(&chan->dev->device,
				"Transfer size must be aligne to %d Bytes, got %d Bytes\n",
				SVEC_DMA_DDR_ALIGN, sg_dma_len(sg));
                        return NULL;
		}
	}

	svec_dma_tx = kzalloc(sizeof(struct svec_dma_tx), GFP_NOWAIT);
	if (!svec_dma_tx)
                return NULL;

	dma_async_tx_descriptor_init(&svec_dma_tx->tx, chan);
	svec_dma_tx->tx.tx_submit = svec_dma_tx_submit;
	svec_dma_tx->sgl = sgl;
	svec_dma_tx->sg_len = sg_len;
        svec_dma_tx->direction = direction;
        svec_dma_tx->vaddr = sconfig->src_addr;

#if 0
	for_each_sg(sgl, sg, sg_len, i) {
		struct gn412x_dma_tx_hw *tx_hw = gn412x_dma_tx->sgl_hw[i];

		dev_dbg(&chan->dev->device,
			"%s\n"
			"\tsegment: %d\n"
			"\tstart_addr: 0x%x\n"
			"\tdma_addr_l: 0x%x\n"
			"\tdma_addr_h: 0x%x\n"
			"\tdma_len: 0x%x\n"
			"\tnext_addr_l: 0x%x\n"
			"\tnext_addr_h: 0x%x\n"
			"\tattribute: 0x%x\n",
			__func__, i,
			tx_hw->start_addr,
			tx_hw->dma_addr_l,
			tx_hw->dma_addr_h,
			tx_hw->dma_len,
			tx_hw->next_addr_l,
			tx_hw->next_addr_h,
			tx_hw->attribute);
	}
#endif

	return &svec_dma_tx->tx;
}

static void svec_dma_submit_sgl(struct svec_fpga *svec_fpga);
static void svec_dma_submit_tx(struct svec_fpga *svec_fpga, struct svec_dma_tx *tx);

static enum dma_status svec_dma_tx_status(struct dma_chan *chan,
                                          dma_cookie_t cookie,
                                          struct dma_tx_state *state)
{
	return DMA_ERROR;
}

static void svec_dma_complete(void *arg, const struct dmaengine_result *result)
{
        struct svec_fpga *svec_fpga = arg;
        struct svec_dma_tx *tx;
        unsigned long flags;

        if (svec_fpga->dma_sg_ptr) {
                /* Continue with the original tx entry */
                svec_dma_submit_sgl(svec_fpga);
                return;
        }

        /* The callback */
        spin_lock_irqsave(&svec_fpga->dma_lock, flags);
        tx = svec_fpga->dma_cur_tx;
        svec_fpga->dma_cur_tx = NULL;
        spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);

        tx->tx.callback_result(tx->tx.callback_param, result);
        kfree(tx);

        /* Next tx entry */
	spin_lock_irqsave(&svec_fpga->dma_lock, flags);
        tx = list_first_entry_or_null(&svec_fpga->dma_pending_list, struct svec_dma_tx, list);
        if (tx) {
                list_del(&tx->list);
        } else {
                dma_release_channel(svec_fpga->vme_chan);
                svec_fpga->vme_chan = NULL;
        }
        spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);

        /* Submit it.  */
        if (tx)
                svec_dma_submit_tx(svec_fpga, tx);
}

static void svec_dma_build_desc(struct svec_fpga *svec_fpga, struct vme_dma *desc, struct svec_dma_tx *tx, unsigned len, dma_addr_t host_addr)
{
        struct vme_dma_attr *vme;
	struct vme_dma_attr *pci;
        dma_addr_t vme_addr;

        vme_addr = svec_fpga->vme_addr;
        switch (tx->tx.chan->chan_id) {
        case 0:
                vme_addr += SVEC_BASE_REGS_DDR4_DATA;
                break;
        case 1:
                vme_addr += SVEC_BASE_REGS_DDR5_DATA;
                break;
        default:
                BUG();
        }

	memset(desc, 0, sizeof(struct vme_dma));

	vme = &desc->src;
	pci = &desc->dst;

	desc->dir	= VME_DMA_FROM_DEVICE;
	desc->length    = len;
	desc->novmeinc  = 1;

	desc->ctrl.pci_block_size   = VME_DMA_BSIZE_4096;
	desc->ctrl.pci_backoff_time = VME_DMA_BACKOFF_0;
	desc->ctrl.vme_block_size   = VME_DMA_BSIZE_4096;
	desc->ctrl.vme_backoff_time = VME_DMA_BACKOFF_0;

	vme->data_width = VME_D32;
	vme->am         = VME_A24_USER_MBLT;
	vme->addru	= 0;
	vme->addrl	= vme_addr;

	pci->addru = upper_32_bits((unsigned long)host_addr);
	pci->addrl = lower_32_bits((unsigned long)host_addr);

}

/* Submit the current scatterlist element to the vme DMA.
 * A new scatterlist is built from the current one with length of each
 * element limited to 4KB so that the VME address stays in the DMA window.
 */
static void svec_dma_submit_sgl(struct svec_fpga *svec_fpga)
{
        struct svec_dma_tx *tx = svec_fpga->dma_cur_tx;
        struct dma_async_tx_descriptor *vtx;
        unsigned i;
        struct scatterlist *sgl;
        unsigned sgl_len;
        unsigned off;
	struct vme_dma desc;
        unsigned total_len;
        dma_addr_t phys_addr;

        sg_init_table(svec_fpga->dma_sg, SVEC_DMA_SG);

        sgl = svec_fpga->dma_sg_ptr;
        off = svec_fpga->dma_sg_off;
        sgl_len = SVEC_DMA_SG;
        total_len = 0;
        phys_addr = sg_phys(sgl);
        for (i = 0; i < SVEC_DMA_SG; i++) {
                struct scatterlist *s = &svec_fpga->dma_sg[i];
                unsigned maxlen = sgl->length - off;
                unsigned doff = sgl->offset + off;
                unsigned pgoff = doff >> PAGE_SHIFT;
                if (maxlen > PAGE_SIZE)
                        maxlen = PAGE_SIZE;

                sg_set_page(s, sg_page(sgl) + pgoff, doff, maxlen);
                sg_dma_len(s) = maxlen;
                sg_dma_address(s) = sg_dma_address(sgl) + off;
                total_len += maxlen;

                off += maxlen;
                if (off >= sgl->length) {
                        /* Next sg entry */
                        off = 0;
                        sgl = sg_next(sgl);
                        svec_fpga->dma_sg_ptr = sgl;
                        if (sgl == NULL) {
                                sg_mark_end(&svec_fpga->dma_sg[i]);
                                sgl_len = i + 1;
                                break;
                        }
                }
        }
        svec_fpga->dma_sg_ptr = sgl;
        svec_fpga->dma_sg_off = off;

        svec_dma_build_desc(svec_fpga, &desc, tx, total_len, phys_addr);

        vtx = svec_fpga->vme_chan->device->device_prep_slave_sg
                (svec_fpga->vme_chan, svec_fpga->dma_sg, sgl_len,
                 tx->direction, 0, &desc);

        if (!vtx) {
                const struct dmaengine_result result = {
                        .result = DMA_TRANS_ABORTED,
                        .residue = 0,
                };
                WARN_ON(1);
                /* Discard current descriptor */
                svec_fpga->dma_sg_ptr = NULL;
                /* Callback */
                svec_dma_complete(svec_fpga, &result);
                return;
        }

        tx->vtx = vtx;
        vtx->callback_result = svec_dma_complete;
        vtx->callback_param = svec_fpga;

        tx->cookie = dmaengine_submit(vtx);
        WARN_ON(tx->cookie < 0);

        dma_async_issue_pending(svec_fpga->vme_chan);

        return;
}

/* Submit the current descriptor to the VME DMA engine. To follow the SVEC
 * DMA mechanism the VME address is first written to the DMA address register,
 * and the VME DMA will be done with the DMA window.
 */
static void svec_dma_submit_tx(struct svec_fpga *svec_fpga, struct svec_dma_tx *tx)
{
        dma_addr_t vaddr;
        unsigned reg_off;
        int err;

        vaddr = svec_fpga->vme_addr;
        switch (tx->tx.chan->chan_id) {
        case 0:
                vaddr += SVEC_BASE_REGS_DDR4_DATA;
                reg_off = SVEC_BASE_REGS_CSR_DDR4_ADDR;
                break;
        case 1:
                vaddr += SVEC_BASE_REGS_DDR5_DATA;
                reg_off = SVEC_BASE_REGS_CSR_DDR5_ADDR;
                break;
        default:
                BUG();
        }

        if (0) {
                /* In fact, this is not supported by vme_dmaengine */
                struct dma_slave_config sconfig;

                sconfig.src_addr = vaddr;
                sconfig.direction = tx->direction;
                sconfig.src_addr_width = 8;

                err = dmaengine_slave_config(svec_fpga->vme_chan, &sconfig);
                WARN_ON(err);
                return;
        }

        /* Write the SVEC dma page address */
        iowrite32be(tx->vaddr, svec_fpga->fpga + reg_off);

        WARN_ON(svec_fpga->dma_cur_tx != NULL);
        svec_fpga->dma_sg_ptr = tx->sgl;
        svec_fpga->dma_cur_tx = tx;

        svec_dma_submit_sgl(svec_fpga);
        return;
}

static bool svec_dmaengine_filter(struct dma_chan *dchan, void *arg)
{
	struct vme_dev *vdev = arg;
        /* The vme board, vme slot, vme bridge  */
	struct device *device_ref = vdev->dev.parent->parent;

	return (dchan->device->dev == device_ref);
}


/* Issue pending requests to the VME DMA engine */
static void svec_dma_issue(struct svec_fpga *svec_fpga)
{
        unsigned long flags;
	dma_cap_mask_t dma_mask;
        struct svec_dma_tx *tx;
        struct svec_dev *svec_dev = to_svec_dev(svec_fpga->dev.parent);
	struct vme_dev *vdev = to_vme_dev(svec_dev->dev.parent);

        /* Allocate channel from vme, unless already allocated */
	spin_lock_irqsave(&svec_fpga->dma_lock, flags);
        if (svec_fpga->vme_chan) {
                spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);
                return;
        }

	dma_cap_zero(dma_mask);
	dma_cap_set(DMA_SLAVE, dma_mask);
	dma_cap_set(DMA_PRIVATE, dma_mask);
	svec_fpga->vme_chan = dma_request_channel(dma_mask,
                                                  svec_dmaengine_filter, vdev);

	if (!svec_fpga->vme_chan) {
                /* Cannot request the DMA channel. We cannot wait here so we
                   need to cancel all the requests. */
                const struct dmaengine_result result = {
                        .result = DMA_TRANS_ABORTED,
                        .residue = 0,
                };
                pr_info("%s: cannot request channel\n",
                        dev_name(&svec_dev->dev));
                while (1) {
                        tx = list_first_entry_or_null(&svec_fpga->dma_pending_list, struct svec_dma_tx, list);
                        if (!tx)
                                break;
                        list_del(&tx->list);
                        tx->tx.callback_result(tx->tx.callback_param, &result);
                        kfree(tx);
                }
                spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);
		return;
        }

        /* Extract entry */
        tx = list_first_entry_or_null(&svec_fpga->dma_pending_list, struct svec_dma_tx, list);
        WARN_ON(tx == NULL);
        list_del(&tx->list);

        spin_unlock_irqrestore(&svec_fpga->dma_lock, flags);

        /* Submit it.  */
        svec_dma_submit_tx(svec_fpga, tx);
}


static void svec_dma_issue_pending(struct dma_chan *dchan)
{
	struct svec_dma_chan *chan = to_svec_dma_chan(dchan);
	struct svec_fpga *svec_fpga = container_of(dchan->device, struct svec_fpga, dma);
        unsigned long flags1;
        unsigned long flags2;

        /* Append the channel tx list to the dma tx list */
	spin_lock_irqsave(&svec_fpga->dma_lock, flags1);
	spin_lock_irqsave(&chan->lock, flags2);
        list_splice_tail_init(&chan->pending_list, &svec_fpga->dma_pending_list);

	spin_unlock_irqrestore(&chan->lock, flags2);
        spin_unlock_irqrestore(&svec_fpga->dma_lock, flags1);

        svec_dma_issue(svec_fpga);
}

#if KERNEL_VERSION(4, 0, 0) > LINUX_VERSION_CODE
static int svec_dma_device_control(struct dma_chan *chan,
                                   enum dma_ctrl_cmd cmd,
                                   unsigned long arg)
{
	switch (cmd) {
	case DMA_SLAVE_CONFIG:
		return svec_dma_slave_config(chan,
                                             (struct dma_slave_config *)arg);
	case DMA_TERMINATE_ALL:
		return svec_dma_terminate_all(chan);
	case DMA_PAUSE:
		break;
	case DMA_RESUME:
		break;
	default:
		break;
	}
	return -ENODEV;
}
#endif

int svec_dma_init(struct svec_fpga *svec_fpga)
{
  	struct dma_device *dma = &svec_fpga->dma;
        struct svec_dev *svec_dev = to_svec_dev(svec_fpga->dev.parent);
	struct vme_dev *vdev = to_vme_dev(svec_dev->dev.parent);
        unsigned i;
        int err;

	dma->dev = &svec_fpga->dev;

	if (dma_set_mask(dma->dev, DMA_BIT_MASK(64))) {
		dev_warn(dma->dev, "64-bit DMA addressing not available\n");

		/* Check if hardware supports 32-bit DMA */
		if (dma_set_mask(dma->dev, DMA_BIT_MASK(32))) {
			dev_err(dma->dev,
				"32-bit DMA addressing not available\n");
			return -EINVAL;
		}
	}

	dma_cap_zero(dma->cap_mask);
	dma_cap_set(DMA_SLAVE, dma->cap_mask);
	dma_cap_set(DMA_PRIVATE, dma->cap_mask);

	dma->device_alloc_chan_resources = svec_dma_alloc_chan_resources;
	dma->device_free_chan_resources = svec_dma_free_chan_resources;
	dma->device_prep_slave_sg = svec_dma_prep_slave_sg;
#if KERNEL_VERSION(4, 0, 0) > LINUX_VERSION_CODE
	dma->device_control = svec_dma_device_control;
#else
	/* TODO: adjust/verify addr widths, direction and granularity. */
	dma->src_addr_widths = DMA_SLAVE_BUSWIDTH_4_BYTES;
	dma->dst_addr_widths = DMA_SLAVE_BUSWIDTH_4_BYTES;
	dma->directions =
		1 << DMA_DEV_TO_MEM |
		1 << DMA_MEM_TO_DEV;
	dma->residue_granularity = 0;
	dma->device_config = svec_dma_slave_config;
	dma->device_terminate_all = svec_dma_terminate_all;
#endif
	dma->device_tx_status = svec_dma_tx_status;
	dma->device_issue_pending = svec_dma_issue_pending;

        INIT_LIST_HEAD(&svec_fpga->dma_pending_list);
        spin_lock_init(&svec_fpga->dma_lock);

	INIT_LIST_HEAD(&dma->channels);
        for (i = 0; i < SVEC_DMA_CHANS; i++) {
                struct svec_dma_chan *chan = &svec_fpga->dma_chan[i];
                chan->chan.device = dma;
                dma_cookie_init(&chan->chan);
                INIT_LIST_HEAD(&chan->pending_list);
                spin_lock_init(&chan->lock);
                list_add_tail(&chan->chan.device_node, &dma->channels);
        }

	dma_set_max_seg_size(dma->dev, 128 << 20);

        svec_fpga->vme_addr = vdev->map[SVEC_FUNC_NR].vme_addrl;

        err = dma_async_device_register(dma);
        return err;
}

int svec_dma_exit(struct svec_fpga *svec_fpga)
{
        unsigned i;

        for (i = 0; i < SVEC_DMA_CHANS; i++)
                dmaengine_terminate_all(&svec_fpga->dma_chan[i].chan);
	dma_async_device_unregister(&svec_fpga->dma);

        return 0;
}
